#!/usr/bin/env python
import re
from collections import defaultdict
import analyser

from find_classes_result import ClassResult
from find_classes_result import SearchResult

##classInfo class
class classInfo:
    def __init__(self, name, line_no, scope):
        self.name = name
        self.line_no = line_no
        self.parents = set()
        #self.objects = set()
        self.children = set()
        self.scope = scope

    def __del__(self):
        pass 
        #print('object deleted')    

    def setParent(self, parent):
        self.parents.add(parent)


def getClass(name, classInfoObj_set):
    for c in classInfoObj_set:
        if c.name == name:
            return c

def getClassWithScope(name, scope, classInfoObj_set):
    for c in classInfoObj_set:
        if c.name == name and c.scope == scope:
            return c
    return None        

def classExistsInScope(name, scope, classInfoObj_set):
    for c in classInfoObj_set:
        if c.name == name and c.scope == scope:
            return True
    return False   

def classExistsOnLine(name, line, classInfoObj_set):
	for c in classInfoObj_set:
		if c.name == name and c.line_no == line:
			return True
	return False        

#dictionaries of tokens we're looking for
comments = {r'//' : 'Single Line Comment',r'/*' : 'Multiline Comment Start', r'*/' : 'Multiline Comment End', '/**/' : 'Empty Multiline comment'}
comment_keys = comments.keys()

encapsulation = {'public' : 'public', 'private' : 'private', 'protected' : 'protected'}
encapsulation_keys = encapsulation.keys()

class_declaration = {'class':'class declaration', 'struct': 'struct declaration', 'union': 'union declaration'}
class_declaration_keys = class_declaration.keys()

blocks = {'{' : 'Blocked Statement Body Open', '}':'Blocked Statement Body Closed'}
block_keys = blocks.keys()


def find_classes(filename):
    print ("__________________________________________________________") 
    print('Running on file: ', filename)

    if analyser.analyse(filename) == 1:
        return 

    classInfoObj_set = set()   
    classes = defaultdict(list)
    forward_classes = defaultdict(list)
    
    
    # Flags
    scope = 0
    comment_flag = 0
    class_flag = 0
    frwd_dec_flag = 0

    count = 0 #counter for line number
    class_name = '' #default class name variable
    frwd_class_name = ''

    f = open(filename,'r') #c++ source file
    src = f.read()  #source text
    program =  src.split('\n')  #text of the source file spiltted line by line


    for line in program:
        count = count+1
        #line = re.sub('::', '_', line)
        
        line = re.sub(':|,', ' ', line) 
        tokens = line.split()
        #tokens = tokens.split(':')   
        
        #print ("Tokens are",tokens)
        #print ("Line #",count,'properties \n')
        #print ('class_flag = ', class_flag)

        if len(tokens) > 0:
            if '//' in tokens[0]:
         #       print ("________________________________________________________________________") 
                continue

        for token in tokens:
            # print("a = ", comment_flag)
            if '//' in token:
                break

            if '\r' in token:
                position = token.find('\r')
                token=token[:position]
            
            if '};' in token and class_flag > 0 and comment_flag == 0:
                #print("b = ", comment_flag)
                class_flag -= 1 
                #print('class_flag- = ', class_flag)

            if token in comment_keys:
                #print ("Comment Type: ", comments[token])
                if token == '/*' and comment_flag == 0:                 		         				 
                    comment_flag = 1
                if token == '*/' and comment_flag == 1:
                    comment_flag = 0     


            if token in class_declaration_keys and comment_flag == 0:
                class_flag += 1 
                for string in tokens:
                    if ('/*' in string and comment_flag == 0):
                        comment_flag = 1
                        break
                    if ('*/' in string and comment_flag == 1):
                        comment_flag = 0    
                    if (string not in encapsulation_keys and string not in block_keys and string not in class_declaration_keys):
                        if("//" in string or comment_flag != 0):
                            #print("Single line comment")
                            break
                            
                        elif comment_flag == 0:
                            #print('adding classes')   
                            class_name = tokens[1]
                            if (';' in class_name) or (len(tokens) == 3  and ';' == tokens[2]):
                                #print('forward declaration')
                                frwd_dec_flag = 1
                                frwd_class_name = class_name 
                                frwd_class_name = re.sub(';|{', '', frwd_class_name)
                                if class_flag == 1:
                                    if classExistsInScope(frwd_class_name,'global',classInfoObj_set) == True:
                                        print('Line: ' + str(count) + ' - class "' + frwd_class_name + '" already declared in this scope')
                                        return
                                    else: 
                                        classInfo_obj = classInfo(frwd_class_name,count,'global')
                                        classInfoObj_set.add(classInfo_obj)
                                        classes[frwd_class_name].append(count)
                                        forward_classes[frwd_class_name].append(count)
                                        class_flag -= 1        
                            
                            if classExistsOnLine(class_name, count, classInfoObj_set) == False: 
                                class_name = re.sub(';|{', '', class_name)
                                #print('normal declaration')
                                if class_flag == 1:
                                    #print('test1')
                                    if classExistsInScope(class_name,'global',classInfoObj_set) == True:
                                        if class_name not in forward_classes.keys():    
                                            print('Line: ' + str(count) + ' - class "' + class_name + '" already declared in this scope')
                                            return
                                    else:
                                        #print('test2') 
                                        classInfo_obj = classInfo(class_name,count,'global')
                                        classInfoObj_set.add(classInfo_obj)
                                        classes[class_name].append(count)
                                if class_flag > 1:
                                        classInfo_obj = classInfo(class_name,count,'local')
                                        classInfoObj_set.add(classInfo_obj)
                                        classes[class_name].append(count)
                            #print(classes)            
                            string = re.sub(',|{|;', '', string)
                            #print('string = ',string)
                            if len(tokens) > 2 and classInfo_obj.name != string and tokens[len(tokens)-1] != '{':
                                c = getClass(string, classInfoObj_set)
                                #print('parent: ', c.name)
                                #print('child: ', classInfo_obj.name)
                                if string not in classes.keys():
                                    print ('Line: ' + str(count) + ' - Class "' + string + '" does not exist')
                                    return
                                if c.name != classInfo_obj.name:
                                    #line_num = count
                                    c_ = getClassWithScope(c.name, 'global', classInfoObj_set)
                                    classInfo_obj.setParent(c_)

               

        #print ("________________________________________________________________________")    

    f.close()

    if class_flag != 0:
        #print(class_flag)
        print('Imbalanced parentheses in class declarations')
        return 

    if len(classes) == 0:
        print('No classes found')
        return


    # print ('classes: ', classes)
    # print('\n')
    print()

    for c in classInfoObj_set:
        print('class "' + c.name + '" declared on line ' + str(c.line_no)) #+ ' scope = ' + c.scope)
        #print(len(c.parents))
        if len(c.parents) > 0:
            print('parent class(es):')
        for i in c.parents:
            print(i.name + ' (Line ' + str(i.line_no) + ')')
        #print ()            
        print()    

    print(classes)
    #classes = 0

    return
	
def find_classes_for_test(filename):
    #print ("__________________________________________________________") 
    #print('Running on file: ', filename)

    if analyser.analyse(filename) == 1:
        return "Error"

    classInfoObj_set = set()   
    classes = defaultdict(list)
    forward_classes = defaultdict(list)

    toReturn = SearchResult()
    
    # Flags
    scope = 0
    comment_flag = 0
    class_flag = 0
    frwd_dec_flag = 0

    count = 0 #counter for line number
    class_name = '' #default class name variable
    frwd_class_name = ''

    f = open(filename,'r') #c++ source file
    src = f.read()  #source text
    program =  src.split('\n')  #text of the source file spiltted line by line


    for line in program:
        count = count+1
        #line = re.sub('::', '_', line)
        
        line = re.sub(':|,', ' ', line) 
        tokens = line.split()
        #tokens = tokens.split(':')   
        
        #print ("Tokens are",tokens)
        #print ("Line #",count,'properties \n')
        #print ('class_flag = ', class_flag)

        if len(tokens) > 0:
            if '//' in tokens[0]:
         #       print ("________________________________________________________________________") 
                continue

        for token in tokens:
            # print("a = ", comment_flag)
            if '//' in token:
                break

            if '\r' in token:
                position = token.find('\r')
                token=token[:position]
            
            if '};' in token and class_flag > 0 and comment_flag == 0:
                #print("b = ", comment_flag)
                class_flag -= 1 
                #print('class_flag- = ', class_flag)

            if token in comment_keys:
                #print ("Comment Type: ", comments[token])
                if token == '/*' and comment_flag == 0:                 		         				 
                    comment_flag = 1
                if token == '*/' and comment_flag == 1:
                    comment_flag = 0     


            if token in class_declaration_keys and comment_flag == 0:
                class_flag += 1 
                for string in tokens:
                    if ('/*' in string and comment_flag == 0):
                        comment_flag = 1
                        break
                    if ('*/' in string and comment_flag == 1):
                        comment_flag = 0    
                    if (string not in encapsulation_keys and string not in block_keys and string not in class_declaration_keys):
                        if("//" in string or comment_flag != 0):
                            #print("Single line comment")
                            break
                            
                        elif comment_flag == 0:
                            #print('adding classes')   
                            class_name = tokens[1]
                            if (';' in class_name) or (len(tokens) == 3  and ';' == tokens[2]):
                                #print('forward declaration')
                                frwd_dec_flag = 1
                                frwd_class_name = class_name 
                                frwd_class_name = re.sub(';|{', '', frwd_class_name)
                                if class_flag == 1:
                                    if classExistsInScope(frwd_class_name,'global',classInfoObj_set) == True:
                                        #print('Line: ' + str(count) + ' - class "' + frwd_class_name + '" already declared in this scope')
                                        tmp = ClassResult()
                                        tmp.className = frwd_class_name
                                        tmp.error_dictionary["already declared in this scope"] = True
                                        toReturn.listOfClasses.append(tmp)
                                        return toReturn
                                    else: 
                                        classInfo_obj = classInfo(frwd_class_name,count,'global')
                                        classInfoObj_set.add(classInfo_obj)
                                        classes[frwd_class_name].append(count)
                                        forward_classes[frwd_class_name].append(count)
                                        class_flag -= 1        
                            
                            if classExistsOnLine(class_name, count, classInfoObj_set) == False: 
                                class_name = re.sub(';|{', '', class_name)
                                #print('normal declaration')
                                if class_flag == 1:
                                    #print('test1')
                                    if classExistsInScope(class_name,'global',classInfoObj_set) == True:
                                        if class_name not in forward_classes.keys():    
                                            #print('Line: ' + str(count) + ' - class "' + class_name + '" already declared in this scope')
                                            tmp = ClassResult()
                                            tmp.className = class_name
                                            tmp.error_dictionary["already declared in this scope"] = True
                                            toReturn.listOfClasses.append(tmp)
                                            return toReturn
                                    else:
                                        #print('test2') 
                                        classInfo_obj = classInfo(class_name,count,'global')
                                        classInfoObj_set.add(classInfo_obj)
                                        classes[class_name].append(count)
                                if class_flag > 1:
                                        classInfo_obj = classInfo(class_name,count,'local')
                                        classInfoObj_set.add(classInfo_obj)
                                        classes[class_name].append(count)
                            #print(classes)            
                            string = re.sub(',|{|;', '', string)
                            #print('string = ',string)
                            if len(tokens) > 2 and classInfo_obj.name != string and tokens[len(tokens)-1] != '{':
                                c = getClass(string, classInfoObj_set)
                                #print('parent: ', c.name)
                                #print('child: ', classInfo_obj.name)
                                if string not in classes.keys():
                                    #print ('Line: ' + str(count) + ' - Class "' + string + '" does not exist')
                                    tmp = ClassResult()
                                    tmp.className = string
                                    tmp.error_dictionary["does not exist"] = True
                                    toReturn.listOfClasses.append(tmp)									
                                    return toReturn
                                if c.name != classInfo_obj.name:
                                    c_ = getClassWithScope(c.name, 'global', classInfoObj_set)
                                    classInfo_obj.setParent(c_)

               

        #print ("________________________________________________________________________")    

    f.close()

    if class_flag != 0:
        #print(class_flag)
#        print('Imbalanced parentheses in class declarations')
        toReturn.error_dictionary_overall["Imbalanced parentheses in class declarations"] = True
        return toReturn

    if len(classes) == 0:
        toReturn.error_dictionary_overall["No classes found"] = True
        #print('No classes found')
        return toReturn
    #print()

    for c in classInfoObj_set:
        nextClass = ClassResult()
        nextClass.className = c.name
        nextClass.classLine = c.line_no
        #print('class "' + c.name + '" declared on line ' + str(c.line_no)) #+ ' scope = ' + c.scope)
        if len(c.parents) > 0:
            pass
            #print('parent class(es):')
        for i in c.parents:
            
            nextClass.parents.append({'name':i.name,'line':i.line_no})
            #print(i.name + ' (Line ' + str(i.line_no) + ')')
        #print()    
        toReturn.listOfClasses.append(nextClass)
    #print(classes)
    #classes = 0

    return toReturn

	
	
#-------------------------------------------------------------------------------------------------------------

list_of_testable_files = []
list_of_testable_files.append('sample.cpp')
list_of_testable_files.append('string.cpp')
list_of_testable_files.append('comments.cpp')
list_of_testable_files.append('wrong_keyword.cpp')
list_of_testable_files.append('nonexist_class.cpp')
list_of_testable_files.append('noClassName.cpp')
list_of_testable_files.append('noColon.cpp')
list_of_testable_files.append('imbalancedParan.cpp')
list_of_testable_files.append('class.cpp')
list_of_testable_files.append('frwd_dec.cpp')


for t in list_of_testable_files:
	pass
#	print('** {0} **'.format(t))
#	print(find_classes_for_test(t))


#--Test cases
#find_classes('sample.cpp')              #contains classes(inlcuding nested) and comments, (10 classes)
#find_classes('string.cpp')              #no classes, string containg "...class..."  
#find_classes('comments.cpp')            #no classes, file with commented out code
#find_classes('nested.cpp')             #4 classes    
#find_classes('wrong_keyword.cpp')       #wrong keyword
#find_classes('nonexist_class.cpp')      #class does not exist
#find_classes('noClassName.cpp')         #incorrect class declaration
#find_classes('noColon.cpp')             #incorrect class declaration
#find_classes('imbalancedParan.cpp')     #imbalanced paranthesis
#find_classes('class.cpp')                 #2 classes in the same scope
#find_classes('frwd_dec.cpp')

