import ast
import copy

try:
    from type_analysis import mark_name,UnknownType
    from operators.base import MutationResign, MutationOperator
except (ImportError,ModuleNotFoundError) as e:
    from mutpy.type_analysis import mark_name,UnknownType
    from mutpy.operators.base import MutationResign, MutationOperator
import pdb

class ConstantReplacementByAVariable(MutationOperator):
	replacement_dictionary = {
		int:ast.Num(1),
		float:ast.Num(1.0),
		complex:ast.Num(1j),
		str:ast.Str("mutpy"),
		list:ast.List(elts=[]),
		tuple:ast.Tuple(),
		range:ast.Call(args=[ast.Num(1)],func = ast.Name('range'),keywords=[]),
		bytes:ast.Bytes(s=bytes()),
		bytearray:ast.Call(args=[],func = ast.Name('bytearray'),keywords=[]),
		memoryview:ast.Call(args=[ast.Num(0)],func = ast.Name('memoryview'),keywords=[]),
		set:ast.Set(elts=[]),
		frozenset:ast.Call(args=[ast.Num(0)],func = ast.Name('frozenset'),keywords=[]),
		dict:ast.Dict(keys=[],values=[]),
		#function:ast.Lambda(args=ast.arguments(args=[],defaults=[],kw_defaults=[],kwarg=[],kwonlyargs=[],vararg=[]),body=ast.NameConstant(value=True)),
		bool:ast.NameConstant(value=True)
	}
	def is_a_replacable_variable(self,node):
		if not hasattr(node,'parent') or not hasattr(node,mark_name):
			return False
		if isinstance(node.parent,ast.Assign) and node.parent.value == node:	#a = b -> a = 0
			return True
		if isinstance(node.parent,ast.BinOp) and (node.parent.right == node or node.parent.left == node):	#a = 1+b -> a = 1+0
			return True
		if isinstance(node.parent,ast.Call) and node in node.parent.args:	#fun(a) -> fun(0)
			return True
		if isinstance(node.parent,ast.ListComp) and node in node.parent.elt:	#[ x for ...] -> [0 for ...]
			return True
		return False
	
	def get_replacement(self,node):
		types = getattr(node,mark_name)
		if len(types) > 1 or UnknownType in types:
			raise MutationResign()
		requested_type = list(types)[0]
		if requested_type in ConstantReplacementByAVariable.replacement_dictionary:
			return copy.copy(ConstantReplacementByAVariable.replacement_dictionary[requested_type])

		raise MutationResign()
	def mutate_Name(self,node):
		#pdb.set_trace()
		if not self.is_a_replacable_variable(node):
			raise MutationResign()
		replacement_node = self.get_replacement(node)
		return replacement_node
		