import ast
import copy

def reduce_duplicates(node,attr_name):
	if not isinstance(node,ast.AST):
		return
	if hasattr(node,attr_name):
		l = list(set(getattr(node,attr_name)))
		setattr(node,attr_name,l)
	if hasattr(node,'_fields'):
		for name in node._fields:
			x = getattr(node,name)
			if isinstance(x,list) or isinstance(x,tuple):
				for a in x:
					reduce_duplicates(a,attr_name)
			else:
				reduce_duplicates(x,attr_name)
	if hasattr(node,'_attributes'):
		for name in node._attributes:
			x = getattr(node,name)
			if isinstance(x,list) or isinstance(x,tuple):
				for a in x:
					reduce_duplicates(a,attr_name)
			else:
				reduce_duplicates(x,attr_name)	

def delete_load_and_store_parents(node):
	if not isinstance(node,ast.AST):
		return
	attr_name = 'parents'
	if (isinstance(node, ast.Load) or isinstance(node, ast.Store)) and hasattr(node,attr_name):
		delattr(node,attr_name)
	else:
		if hasattr(node,'_fields'):
			for name in node._fields:
				x = getattr(node,name)
				if isinstance(x,list) or isinstance(x,tuple):
					for a in x:
						delete_load_and_store_parents(a)
				else:
					delete_load_and_store_parents(x)
		if hasattr(node,'_attributes'):
			for name in node._attributes:
				x = getattr(node,name)
				if isinstance(x,list) or isinstance(x,tuple):
					for a in x:
						delete_load_and_store_parents(a)
				else:
					delete_load_and_store_parents(x)					

def delete_children_and_parents_attributes(node,parent):
	if not isinstance(node,ast.AST):
		return
	if hasattr(node,'parents'):
		delattr(node,'parents')
	if hasattr(node,'children'):
		delattr(node,'children')
	#if not parent is None:
	#	node.parent = parent	
		
	if hasattr(node,'_fields'):
		for name in node._fields:
			x = getattr(node,name,None)
			if not x is None:
				if isinstance(x,list) or isinstance(x,tuple):
					for a in x:
						delete_children_and_parents_attributes(a,node)
				else:
					delete_children_and_parents_attributes(x,node)
	if hasattr(node,'_attributes'):
		for name in node._attributes:
			x = getattr(node,name,None)
			if not x is None:
				if isinstance(x,list) or isinstance(x,tuple):
					for a in x:
						delete_children_and_parents_attributes(a,node)
				else:
					delete_children_and_parents_attributes(x,node)


def delete_attributes(node,attr_name_list,parent):
	if not isinstance(node,ast.AST):
		return
	for attr_name in attr_name_list:
		if hasattr(node,attr_name):
			delattr(node,attr_name)
	#if not parent is None:
	#	node.parent = parent	
	l = []
	if hasattr(node,'_fields'):
		l.extend(node._fields)
	if hasattr(node,'_attributes'):
		l.extend(node._attributes)
	for name in l:
		x = getattr(node,name,None)
		if not x is None:
			if isinstance(x,list) or isinstance(x,tuple):
				for a in x:
					delete_attributes(a,attr_name_list,node)
			else:
				delete_attributes(x,attr_name_list,node)
	

def delete_all_attributes(node,attr_name_list):
	#root = node
	#while hasattr(root,'parent'):
	#	root = root.parent
	parent = getattr(node,'parent',None)
	delete_attributes(node,attr_name_list,parent)
	
def show_elements(node):l = {x:getattr(node,x) for x in node._fields};l['parent'] = getattr(node,'parent','doesn\'t exist');l['type_mark']= getattr(node,'type_mark','doesn\'t exist');return l;

def parents_are_ok(node):all = set(node._fields)|set(node._attributes);children = {getattr(node,x) for x in all if hasattr(getattr(node,x),'__module__')}; tmp = [getattr(node,x) for x in all if hasattr(getattr(node,x),'__iter__') and not isinstance(getattr(node,x),str)]; tmp2 = [0 for x in tmp if children.update(set(x))];d = {x:id(x.parent) == id(node) for x in children};return d;

def all_parents_are_ok(node,depth):
	d = parents_are_ok(node)
	if depth > 0:
		d2 = {}
		for key,value in d.items():
			d2.update(all_parents_are_ok(value,depth-1))
		d.update(d2)
	return {x for x,v in d.items() if not v}
	
	

class RemoveChildrenAndParentsAttributes(ast.NodeTransformer):
	def visit(self,node):
		if hasattr(node,'children'):
			delattr(node,'children')
		if hasattr(node,'parents'):
			delattr(node,'parents')
		result_node = super().visit(node)
		return result_node
		


def detect_no_lineno(node):
	root = node
	while hasattr(root,'parent') and not getattr(root,'parent') is None:
		root = root.parent
	if not isinstance(root,ast.Module):
		return False,'lacks root as ast.Module'
	nodes_that_dont_have_lineno = {ast.Module,ast.Not,ast.Load,ast.Store,ast.Del,ast.In,ast.NotIn,ast.Index,ast.Add,ast.Sub,ast.Div,ast.Pow,ast.FloorDiv,ast.Mod,ast.Mult,ast.Or,ast.And,ast.Lt,ast.Gt,ast.Eq,ast.LtE,ast.GtE,ast.NotEq,ast.BitOr,ast.BitAnd,ast.RShift,ast.LShift,ast.arguments,ast.withitem,ast.keyword,ast.Is,ast.IsNot,ast.comprehension,ast.alias,ast.UAdd,ast.USub,ast.Slice}
	def check_is_lineno(node):
		if not isinstance(node,ast.AST):
			return True, "" ,None
		if not hasattr(node,'lineno') and not type(node) in nodes_that_dont_have_lineno:
			return False,"{}".format(node),node
		l = []
		if hasattr(node,'_fields'):
			l.extend(node._fields)
		if hasattr(node,'_attributes'):
			l.extend(node._attributes)

		for name in l:
			x = getattr(node,name,None)
			if not x is None:
				if isinstance(x,list) or isinstance(x,tuple):
					i=0
					for a in x:
						returned,lead,n = check_is_lineno(a)
						if returned==False:
							return False,"{0}[{1}] {2}".format(name,i,lead),n
						i+=1
				else:
					returned,lead,n = check_is_lineno(x)
					if returned==False:
						return False,"{0} {1}".format(name,lead),n
		return True,"",None
	return check_is_lineno(root)
		
	